from yta_constants.enum import YTAEnum as Enum
from typing import Union


class TimeSpace(Enum):
    """
    The time space that is being used to handle the time
    moment in which something must be applied in the
    editor.
    """

    LOCAL_TIME_NORMALIZED = 'local_normalized'
    """
    The local time normalized, relative to the element.
    """
    LOCAL_TIME = 'local_time'
    """
    The local time (not normalized) relative to the
    element.
    """
    LOCAL_FRAME_INDEXES = 'local_frames'
    """
    The frame indexes, relative to the element.
    """
    # GLOBAL = 'global'
    # """
    # The global time moment, relative to the whole timeline.
    # """

    def to_effective_time_range(
        self,
        start: Union[float, int],
        end: Union[float, int]
    ) -> Union['LocalTimeNormalizedEffectiveTimeRange', 'LocalTimeEffectiveTimeRange', 'LocalFrameIndexesEffectiveTimeRange']:
        """
        Get an `EffectiveTimeRange` instance according to this
        `TimeSpace` enum option value, initialized with the
        `start` and `end` provided.
        """
        return TimeSpace.get_effective_time_range_class(
            start = start,
            end = end
        )

    @staticmethod
    def get_effective_time_range_class(
        time_space: 'TimeSpace'
    ) -> Union['LocalTimeNormalizedEffectiveTimeRange', 'LocalTimeEffectiveTimeRange', 'LocalFrameIndexesEffectiveTimeRange']:
        """
        Get the effective time range class associated to the
        given `time_space`.
        """
        from yta_editor_time.effective_time_range import LocalTimeNormalizedEffectiveTimeRange, LocalTimeEffectiveTimeRange, LocalFrameIndexesEffectiveTimeRange

        return (
            LocalTimeNormalizedEffectiveTimeRange
            if time_space == TimeSpace.LOCAL_TIME_NORMALIZED else
            LocalTimeEffectiveTimeRange
            if time_space == TimeSpace.LOCAL_TIME else
            LocalFrameIndexesEffectiveTimeRange
            if time_space == TimeSpace.LOCAL_FRAME_INDEXES else
            # None
            # if self.space == TimeSpace.GLOBAL else
            None
        )

class TimeAnchor(Enum):
    """
    The part of the element this time specification is
    anchored to and must be considered when calculating
    according to the context given.
    """

    START = 'start'
    """
    Anchored to the `start` of the element and depending
    on it.
    """
    END = 'end'
    """
    Anchored to the `end` of the element and depending on
    it.
    """

"""
Some examples below:
1. Normalized range
    t_start = 0.0
    t_end   = 1.0
    space = LOCAL_TIME_NORMALIZED
    anchor = START

2. Last N seconds
    duration = 0.5
    space = ITEM_SECONDS
    anchor = END

3. Frames
    frame_start = 120
    frame_end   = 180
    space = LOCAL_FRAME_INDEXES
    anchor = START
"""


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class PlaybackDirectionTemporalSpecification(TemporalSpecification):
    """
    TODO: Explain better
    """

    def __init__(
        self,
        do_forward: bool = True
    ):
        self.do_forward: bool = do_forward
        """
        Boolean flag to indicate if it is going forward
        or not (backwards).
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        # Requires ctx.playback_direction
        return TemporalResolution(True)


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class NormalizedRampTemporalSpecification(TemporalSpecification):
    """
    TODO: Explain this better.
    """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        t_local_normalized = evaluation_context.t_normalized
        weight = max(0.0, min(1.0, 1.0 - abs(2.0 * t_local_normalized - 1.0)))

        return TemporalResolution(
            is_active = True,
            t_local_normalized = t_local_normalized,
            weight = weight
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class GlobalNormalizedTimeTemporalSpecification(TemporalSpecification):
    """
    The element is active during a period of time that
    is in between the `start` and `end` normalized global
    t time moment (or progress).

    TODO: Explain better
    """

    def __init__(
        self,
        start: float,
        end: float
    ):
        # TODO: Validate 'start' and 'end' as normalized
        assert start <= end
        self.start: float = start
        self.end: float = end

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        # TODO: This could be handled dynamically
        domain_min = 0.0
        domain_max = 1.0

        start = max(self.start, domain_min)
        end = min(self.end, domain_max)

        t = evaluation_context.t_global_normalized

        if (
            t < start or
            t > end
        ):
            return TemporalResolution(
                is_active = False
            )
        
        t_local_normalized = (t - self.start) / (self.end - self.start)

        return TemporalResolution(
            is_active = True,
            t_local_normalized = t_local_normalized,
            weight = 1.0
        )



from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class BlendTemporalSpecification(TemporalSpecification):
    """
    TODO: Explain better
    """

    def __init__(
        self,
        temporal_specifications: list[TemporalSpecification]
    ):
        self.temporal_specifications: list[TemporalSpecification] = list(temporal_specifications)
        """
        The list of `TemporalSpecification` instances that will
        be evaluated and blended.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        weight = 0.0

        for temporal_specification in self.temporal_specifications:
            temporal_result = temporal_specification.resolve(evaluation_context)

            if temporal_result.active:
                weight += temporal_result.weight

        return TemporalResolution(
            is_active = weight > 0.0,
            t_local_normalized = evaluation_context.t_normalized,
            weight = min(1.0, weight)
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class WeightTemporalSpecification(TemporalSpecification):
    """
    The element is active according to the specific
    and custom function that has been provided to
    this class, that is evaluated according to the
    evaluation context.

    This will always active and only calculate the
    weight based on the `function` provided.
    """

    def __init__(
        self,
        function: callable
    ):
        self.function: callable = function
        """
        The function that must be used to resolve the
        weight based on the evaluation context.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        return TemporalResolution(
            is_active = True,
            t_local_normalized = evaluation_context.t_normalized,
            weight = max(0.0, min(1.0, self.function(evaluation_context)))
        )



from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class FrameTemporalSpecification(TemporalSpecification):
    """
    The element is active during a certain number of
    frames that are in between the `start` and `end`
    frame indexes provided.
    """

    def __init__(
        self,
        start: int,
        end: int
    ):
        # TODO: Validate 'start' and 'end' as normalized
        assert start <= end
        self.start: int = start
        self.end: int = end

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        frame_index = evaluation_context.frame_index
        if (
            frame_index < self.start or
            frame_index > self.end
        ):
            return TemporalResolution(
                is_active = False,
                t_local_normalized = None,
                weight = 1.0
            )

        t_local = (frame_index - self.start) / max(1, self.end - self.start)

        return TemporalResolution(
            is_active = True,
            t_local_normalized = t_local,
            weight = 1.0
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class FirstFrameTemporalSpecification(TemporalSpecification):
    """
    The element is active only if the `frame_index`
    provided in the `evaluation_context` is the first
    frame.
    """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        return TemporalResolution(
            is_active = evaluation_context.frame_index == 0,
            # Use 't_local_normalized' as in the context
            t_local_normalized = None,
            weight = 1.0
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


# TODO: Maybe rename to 'instant' or 'frame' (?)
class LocalPointTemporalSpecification(TemporalSpecification):
    """
    The element is active during a single frame, based
    on the `point` provided (which is a not normalized
    time moment) and with the `epsilon` as the margin
    of error.
    """

    def __init__(
        self,
        point: float,
        epsilon: float = 1e-3
    ):
        self.point: float = point
        self.epsilon: float = epsilon

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        return (
            TemporalResolution(
                is_active = True,
                t_local_normalized = 0.0,
                weight = 1.0
            )
            if abs(evaluation_context.t - self.point) <= self.epsilon else
            TemporalResolution(
                is_active = False,
                t_local_normalized = None,
                weight = 1.0
            )
        )


from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution
from yta_editor_time.enum import TimeSpace, TimeAnchor
from dataclasses import dataclass
from typing import Union


@dataclass(frozen = True)
class TimeRangeSpecification:
    """
    A time range specification in which an element
    should be applied.
    """

    def __init__(
        self,
        space: TimeSpace,
        anchor: TimeAnchor,
        start: Union[int, float, None],
        end: Union[int, float, None],
        duration: Union[float, None]
    ):
        space = TimeSpace.to_enum(space)
        anchor = TimeAnchor.to_enum(anchor)

        self.space: TimeSpace = space
        """
        The time space in which we are specificating the
        time range to be applied.
        """
        self.anchor: TimeAnchor = anchor
        """
        The position, element or part of it to which this
        specification is anchored.
        """
        self.start: Union[int, float, None] = start
        """
        The start time moment of this time specification.
        """
        self.end: Union[int, float, None] = end
        """
        The end time moment of this time specification.
        """
        self.duration: Union[float, None] = duration
        """
        The total duration of this time specification.
        """

    def is_active(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ) -> bool:
        """
        Check if the element with this `TimeRangeSpecification`
        is active or not according to the `evaluation_context`
        provided.
        """
        return self.space.to_effective_time_range(
            start = self.start,
            end = self.end
        ).is_active(evaluation_context)
    
    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        """
        Get the `start` and `end` time moments of this time
        range specification by resolving it according to the
        `evaluation_context` provided.
        """
        if not self.is_active(evaluation_context):
            return TemporalResolution(
                is_active = False,
                t_local_normalized = None,
                weight = 1.0
            )

        # TODO: Resolving according to the type
        if self.anchor == TimeSpace.LOCAL_TIME_NORMALIZED:
            if self.anchor == TimeAnchor.START:
                start = (
                    self.start or
                    0.0
                )
                end = (
                    self.end or
                    start + self.duration
                )
            elif self.anchor == TimeAnchor.END:
                end = 1.0 - (
                        self.start or
                        0.0
                    )
                start = end - self.duration
        elif self.anchor == TimeSpace.LOCAL_TIME:
            if self.anchor == TimeAnchor.START:
                start = (
                    self.start or
                    0.0
                )
                end = (
                    self.end or
                    start + self.duration
                )
            elif self.anchor == TimeAnchor.END:
                end = evaluation_context.item_duration - (
                        self.start or
                        0.0
                    )
                start = end - self.duration
        elif self.anchor == TimeSpace.LOCAL_FRAME_INDEXES:
            if self.anchor == TimeAnchor.START:
                start = (
                    self.start or
                    0
                )
                end = (
                    self.end or
                    # TODO: Is this value the duration in frames or
                    # the total amount of frames (?)
                    start + self.duration_frames
                )
            elif self.anchor == TimeAnchor.END:
                end = evaluation_context.number_of_frames - 1 - (
                        self.start or
                        0
                    )
                start = end - self.duration_frames
        else:
            raise Exception('Something went wrong...')
        # elif self.anchor == TimeSpace.GLOBAL:
        #     if self.anchor == TimeAnchor.START:
        #         start = (
        #             self.start or
        #             0.0
        #         )
        #         end = (
        #             self.end or
        #             start + self.duration
        #         )
        #     elif self.anchor == TimeAnchor.END:
        #         end = evaluation_context.item_duration - (
        #                 self.start or
        #                 0.0
        #             )
        #         start = end - self.duration 

        return TemporalResolution(
            is_active = True,
            t_local_normalized = evaluation_context.t_normalized,
            weight = 1.0
        ) 


        t = evaluation_context.t_normalized

        # TODO: I think this conditions vary depending on the
        # TimeSpace and TimeAnchor
        if (
            t < self.start or
            t > self.end
        ):
            return TemporalResolution(
                is_active = False,
                t_local_normalized = None,
                weight = 1.0
            )

        # TODO: This calculation is used again and again...
        t_local_normalized = (t - self.start) / (self.end - self.start)

        return TemporalResolution(
            is_active = True,
            t_local_normalized = t_local_normalized,
            weight = 1.0
        )


"""
# Todo el clip
TimeRangeSpec(
    space=ITEM_NORMALIZED,
    anchor=START,
    start=0.0,
    end=1.0
)

# Últimos 0.5s
TimeRangeSpec(
    space=ITEM_SECONDS,
    anchor=END,
    duration=0.5
)
"""

"""
| Dominio temporal               | `domain.min` | `domain.max`           |
| ------------------------------ | ------------ | ---------------------- |
| `t_normalized` (local clip)    | `0.0`        | `1.0`                  |
| `t_global_normalized`          | `0.0`        | `1.0`                  |
| `t` (tiempo local en segundos) | `0.0`        | `clip_duration`        |
| `t_global` (timeline)          | `0.0`        | `timeline_duration`    |
| `frame_index`                  | `0`          | `number_of_frames - 1` |

"""


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class OrTemporalSpecification(TemporalSpecification):
    """
    Special `TemporalSpecification`.

    The set of specifications provided will be resolved
    according to the evaluation context. It will be
    active only if at least one of the specifications is
    is_active.

    The `t_local_normalized` value used will be the one
    obtained in the first `TemporalSpecification` resolved
    as active.
    """

    def __init__(
        self,
        temporal_specifications: list[TemporalSpecification]
    ):
        self.temporal_specifications: list[TemporalSpecification] = list(temporal_specifications)
        """
        The list of `TemporalSpecification` instances that will
        be evaluated. All of them must be active to turn the
        final result as active.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        for temporal_specification in self.temporal_specifications:
            temporal_result = temporal_specification.resolve(evaluation_context)

            if temporal_result.is_active:
                return temporal_result
            
        return TemporalResolution(
            is_active = False,
            t_local_normalized = None,
            weight = 1.0
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class NeverTemporalSpecification(TemporalSpecification):
    """
    A `TemporalSpecification` class that indicates that it
    is never active and happening.
    """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        return TemporalResolution(
            is_active = False,
            t_local_normalized = None,
            weight = 1.0
        )



from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class LocalTimeTemporalSpecification(TemporalSpecification):
    """
    The element is active during a period of time that
    is in between the `start` and `end` not normalized
    local t time moment.

    TODO: Explain better
    """

    def __init__(
        self,
        start: float,
        end: float
    ):
        # TODO: Validate 'start' and 'end' as normalized
        assert start <= end
        self.start: float = start
        self.end: float = end

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        t = evaluation_context.t

        if (
            t < self.start or
            t > self.end
        ):
            return TemporalResolution(
                is_active = False,
                t_local_normalized = None,
                weight = 1.0
            )

        t_local_normalized = (t - self.start) / (self.end - self.start)

        return TemporalResolution(
            is_active = True,
            t_local_normalized = t_local_normalized,
            weight = 1.0
        )



from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class BooleanConditionTemporalSpecification(TemporalSpecification):
    """
    The element is active according to the specific
    and custom condition that has been provided.
    """

    def __init__(
        self,
        condition: any
    ):
        self.condition: any = condition
        """
        The condition that must be used to resolve if it
        must be active or not, based on the evaluation
        context.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        return TemporalResolution(
            is_active = bool(self.condition(evaluation_context)),
            # Use 't_local_normalized' as in the context
            t_local_normalized = None,
            weight = 1.0
        )



from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class NotTemporalSpecification(TemporalSpecification):
    """
    Special `TemporalSpecification`.

    The specification provided will be resolved according
    to the evaluation context and this result will be
    active only if the provided one is resolved as non
    active.
    """

    def __init__(
        self,
        temporal_specification: TemporalSpecification
    ):
        self.temporal_specification: TemporalSpecification = temporal_specification
        """
        A `TemporalSpecification` instance that will be 
        evaluated.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        temporal_result = self.temporal_specification.resolve(evaluation_context)

        return TemporalResolution(
            is_active = not temporal_result.is_active,
            # Use 't_local_normalized' as in the context
            t_local_normalized = None,
            weight = 1.0
        )


from dataclasses import dataclass
from typing import Union


@dataclass(frozen = True)
class TemporalResolution:
    """
    Result of resolving a `TemporalSpecification` for a
    given context.

    If the `t_local_normalized` parameter has a value
    means that this is the value that must be used, but
    if it is `None`, the original one provided in the
    context must be used. This is very important because
    we can analyze different specs and setting it wrongly
    could make the next ones think that we are in a new
    temporal subspace.
    """

    def __init__(
        self,
        is_active: bool,
        t_local_normalized: Union[float, None] = None,
        weight: float = 1.0
    ):
        self.is_active: bool = is_active
        """
        Boolean flag to indicate if the element is active or
        not at that moment.
        """
        self.t_local_normalized: Union[float, None] = t_local_normalized
        """
        The local and normalized time moment in which it is
        being analyzed.

        The `None` value means that it must not be redefined
        and the original value must be used.
        """
        self.weight: float = weight
        """
        TODO: Explain it. I don't know what this is for (?)
        """



from abc import ABC, abstractmethod


class EffectiveTimeRange(ABC):
    """
    *Abstract class*

    A time range specification that is able to check
    if the values provided in a context are included
    or not in that time range.

    This class must be have specific implementations
    according to the time space.
    """

    @abstractmethod
    def is_active(
        self,
        evaluation_context: 'VideoEditorEvaluationContext',
        do_include_end: bool = False
    ) -> bool:
        """
        Check if the `evaluation_context` is included or not
        in this time range, including the `end` only if the
        `do_include_end` boolean flag is `True`.
        """
        pass

"""
Specific implementations below
"""
class LocalTimeNormalizedEffectiveTimeRange(EffectiveTimeRange):
    """
    An effective time range that is using a local time
    normalized space. The `start` and `end` values are
    normalized values in the `[0.0, 1.0]` range.
    """

    def __init__(
        self,
        start: float,
        end: float  
    ):
        self.start: float = start
        """
        The normalized local time moment in which the time range
        starts.
        """
        self.end: float = end
        """
        The normalized local time moment in which the time range
        ends.
        """

    def is_active(
        self,
        evaluation_context: 'VideoEditorEvaluationContext',
        do_include_end: bool = False
    ) -> bool:
        """
        Check if the `evaluation_context` is included or not
        in this time range, including the `end` only if the
        `do_include_end` boolean flag is `True`.

        This method will compare the `t_normalized` value of
        the `evaluation_context` provided.
        """
        return (
            self.start <= evaluation_context.t_normalized <= self.end
            if do_include_end else
            self.start <= evaluation_context.t_normalized < self.end
        )
    
class LocalTimeEffectiveTimeRange(EffectiveTimeRange):
    """
    An effective time range that is using a local time
    space. The `start` and `end` values are not normalized
    values.
    """

    def __init__(
        self,
        start: float,
        end: float  
    ):
        self.start: float = start
        """
        The not normalized local time moment in which the time range
        starts.
        """
        self.end: float = end
        """
        The not normalized local time moment in which the time range
        ends.
        """

    def is_active(
        self,
        evaluation_context: 'VideoEditorEvaluationContext',
        do_include_end: bool = False
    ) -> bool:
        """
        Check if the `evaluation_context` is included or not
        in this time range, including the `end` only if the
        `do_include_end` boolean flag is `True`.

        This method will compare the `t` value of the
        `evaluation_context` provided.
        """
        return (
            self.start <= evaluation_context.t <= self.end
            if do_include_end else
            self.start <= evaluation_context.t < self.end
        )
    
class LocalFrameIndexesEffectiveTimeRange(EffectiveTimeRange):
    """
    An effective time range that is using a frames time
    space. The `start` and `end` values are frame indexes.
    """

    def __init__(
        self,
        start: int,
        end: int  
    ):
        self.start: int = start
        """
        The frame index in which the time range starts.
        """
        self.end: int = end
        """
        The frame index in which the time range ends.
        """

    def is_active(
        self,
        evaluation_context: 'VideoEditorEvaluationContext',
        do_include_end: bool = False
    ) -> bool:
        """
        Check if the `evaluation_context` is included or not
        in this time range, including the `end` only if the
        `do_include_end` boolean flag is `True`.

        This method will compare the `frame_index` value of
        the `evaluation_context` provided.
        """
        return (
            self.start <= evaluation_context.frame_index <= self.end
            if do_include_end else
            self.start <= evaluation_context.frame_index < self.end
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class EveryNthFrameTemporalSpecification(TemporalSpecification):
    """
    The element is active during every 'n' number
    of frames.
    """

    def __init__(
        self,
        n: int
    ):
        self.n: int = max(1, n)
        """
        The amount of frames in between one activation of
        this element and the next one.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        is_active = (evaluation_context.frame_index % self.n) == 0

        return TemporalResolution(
            is_active = is_active,
            t_local_normalized = (
                evaluation_context.t_normalized
                if is_active else
                None
            ),
            weight = 1.0
        )


from yta_editor_time.temporal_specification.abstract import TemporalSpecification
from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution


class AndTemporalSpecification(TemporalSpecification):
    """
    Special `TemporalSpecification`.

    The set of specifications provided will be resolved
    according to the evaluation context. It will be
    active only if all of the specifications are active.

    The `t_local_normalized` value used will be the one
    obtained in the last `TemporalSpecification` resolved,
    only if all of them are active.
    """

    def __init__(
        self,
        temporal_specifications: list[TemporalSpecification]
    ):
        self.temporal_specifications: list[TemporalSpecification] = list(temporal_specifications)
        """
        The list of `TemporalSpecification` instances that will
        be evaluated. All of them must be active to turn the
        final result as active.
        """

    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ):
        temporal_specifications_values = []
        weight = 1.0

        for temporal_specification in self.temporal_specifications:
            temporal_result = temporal_specification.resolve(evaluation_context)

            if not temporal_result.is_active:
                return TemporalResolution(
                    is_active = False,
                    t_local_normalized = None,
                    weight = 1.0
                )
            
            # TODO: I think this condition is not needed, all
            # the actives come with the 't_local_normalized'
            if temporal_result.t_local_normalized is not None:
                temporal_specifications_values.append(temporal_result.t_local_normalized)

            weight *= temporal_result.weight

        t_local_normalized = (
            temporal_specifications_values[-1]
            if temporal_specifications_values else
            evaluation_context.t_normalized
        )

        return TemporalResolution(
            is_active = True,
            t_local_normalized = t_local_normalized,
            weight = weight
        )



from yta_editor_time.temporal_specification.temporal_resolution import TemporalResolution
from abc import ABC, abstractmethod


class TemporalSpecification(ABC):
    """
    Abstract base class that defines WHEN something is
    active in a given temporal context.
    """

    @abstractmethod
    def resolve(
        self,
        evaluation_context: 'VideoEditorEvaluationContext'
    ) -> TemporalResolution:
        """
        Resolve this temporal specification for the given
        `evaluation_context`.

        Must be:
        - Pure (no side effects)
        - Deterministic
        - Stateless (resolution is NOT stored)
        """
        pass
